#include "config.h"
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "ports.h"

#define pbMASK   BIN(0001,1111)
#define pdMASK   BIN(1110,1111)

#define LCD_DATA_TO_CLOCK_RISE    100
#define LCD_CLOCK_HOLD_HIGH       100

static void Linitpins()
{
  SETOUT(LCD_E);   PORCLR(LCD_E);
  SETOUT(LCD_RS);  PORCLR(LCD_RS);

  DDRB |= pbMASK;  PORTB &= ~pbMASK;
  DDRD |= pdMASK;  PORTD &= ~pdMASK;
}

static void Lput(uint8_t RS, uint8_t B, uint8_t D)
{
  PORTB = (PORTB & ~pbMASK) | (B&pbMASK);
  PORTD = (PORTD & ~pdMASK) | (B&pdMASK);
  if (RS) PORSET(LCD_RS); else PORCLR(LCD_RS);
  _delay_us(LCD_DATA_TO_CLOCK_RISE);
  PORSET(LCD_E);
  _delay_us(LCD_CLOCK_HOLD_HIGH);
  PORCLR(LCD_E);
  if (D) _delay_ms(5);
    else _delay_us(120);
}

static void Lclear()
{
  Lput(0, BIN(0000,0001), 1); 
}

static void Lgoto(uint8_t r,uint8_t c)
{
  uint8_t a;
  a= 0;
  if (r&1) a|= 0x40;
  if (c>16) c%= 16;
  a+= c;
  Lput(0, BIN(1000,0000) | a, 0);
}

static void Lprt(uint8_t B)
{
  Lput(1, B, 0);
}

static void Linit()
{
  Lput(0, BIN(0011,1100),0); // Function set 8 bits, 2 lines, big font
  Lput(0, BIN(0000,0110),0); // Entry mode, cursor moves right, no display shift
  Lput(0, BIN(0000,1100),0); // Display on, cursor off, blink off
  Lclear();
}

static void str(char *k)
{
  while(*k) Lprt((uint8_t) *(k++));
}


static void Sinit()
{
  UBRRH= USART_DIVIDER>>8;
  UBRRL= USART_DIVIDER;
  UCSRC= (1<<URSEL) | (3<<UCSZ0);
  UCSRB= (1<<RXEN) | (1<<RXCIE);
}

static volatile uint8_t usbuf[BUSIZ], ush, ust;

ISR(USART_RXC_vect)
{
  uint8_t D;
  uint8_t K;
  D= UDR;
  K= (ust+1)%BUSIZ;
  if (K!=ush) 
  {
    usbuf[ust]= D;
    ust= K;
  }
}

static void process()
{
  uint8_t B;
  B= usbuf[ush];
  ush= (ush+1)%BUSIZ;
  switch ((B&BIN(1100,0000))>>6)
  {
  case 0: 
  case 1: Lprt(B); break; 
    // goto command
  case 2: Lgoto( (B >> 4)&1, B&0xf); break;
    // general commands, there is only one for now
  case 3: Lclear(); break;
  }
}

int main()
{
  _delay_ms(10);
  Linitpins();
  _delay_ms(10);
  Linit();
  //   0123456789ABCDEF
  str("1Mhz       12.5k");
  Lgoto(1,0);
  str("UBRR=4s      8N1");
  _delay_ms(1000);
  Sinit();
  sei();

  while(1)
  {
    if (ust!=ush) process();
  }
}


